#!/usr/bin/perl -w
#############################################################################
# Scripture Link Builder        Version 1.6b                                #
# Copyright 1998-9 Ryan Henrie  ryan@ldsinfobase.net                        #
# Created 08/22/98              Last Modified 11/19/06                      #
# Website address:              http://www.ldsinfobase.net/                 #
#############################################################################
# COPYRIGHT NOTICE                                                          #
# Copyright  1998-2004 Ryan C. Henrie  All Rights Reserved.                #
#                                                                           #
# Purpose: This program works on Folio Flat Files (FFF).  When run, this    #
#          script will add the necessary code to make (almost) any scriptre #
#          reference a hyperlink in the whole infobase.  The links are      #
#          formatted to work with The LDS Quad, found on my website.        #
#                                                                           #
# The Scripture Link Builder may be used and modified free of charge by     #
# anyone so long as this copyright notice and these header comments remain  #
# intact.  By using this code you agree to indemnify Ryan Henrie from any   #
# liability that might arise from it's use.                                 #
#                                                                           #
# Use of this program for commercial projects without prior written consent #
# is expressly forbidden. In other words, please ask first before you try   #
# and make money off of my program.                                         #
#                                                                           #
# Obtain permission before redistributing this software over the Internet or#
# in any other medium. In all cases copyright and header must remain intact.#
#                                                                           #
# One case that still could/needs to be handled:                            #
#     A check for or clearing of pre-existing query links, so if you re-run #
#     the script on an infobase, it won't double-up the references.         #
#                                                                           #
# For the ideal script, everything in this list should be parsed correctly: #
#   for simple functionality           John 3:5                             #
#   for multi-verses                   Matt. 5:4-6                          #
#   for complex verses                 gen 4:3-7, 15-16, 19                 #
#   for I John                         I John 2:5-10; exodus 1:1            #
#   for multi-chapter refs             luke 5:2-6:4                         #
#   for whole chapters                 Mark 2; 3.  Gen 2:3; 3; 4:3-9        #
#   for the space after chap #         I think Matt. 5 is good.             #
#   for abbreviated verses             D&C 88:105-12; Deut 10:10-5          #
#   For multi- whole-chapter refs      Job 1-3 is great.                    #
#   for complex multi-chapter refs     Mosiah 8-11, 13, 17 is great         #
#   for comma-not semi-colon-splits    I like it (John 5:3, 6:1)            #
#   For single-chapter books           Omni 3                               #
#                                                                           #
# If anyone wants to add more functionality to this, please send me a copy. #
# If you aren't familiar with Perl, I'm not going to answer questions about #
# how to get this script to run, so you're on your own there.               #
#                                                                           #
# Command line arguments: [input filename] [output filename]                #
#      If no filename is given, it will prompt you for the input filename   #
#      if no output filename is given, it dumps the output to STDOUT.       #
#############################################################################
# Revision History:                                                         #
# v1.7:  Added support for <EN links instead of <QL links                   #
# v1.6b: Added support for links like 2:4-7:28                              #
# v1.51: Added chapter checking to see if the chapter found is logically    #
#            in the real range of chapters for that book. (ie no Luke 60)   #
# v1.5:  Added complex multi- whole chapter queries, and also catch extra   #
#             space at the end of a link better.                            #
#        now it will fix the ocr error of John 5:3, 6:2 OCR error           #
# v1.4:  Working on: getting it to clear any other references before adding #
#             any new ones.                                                 #
#        Updated the tick() subroutine for general purpose use in scripts.  #
# v1.3:  Got multi- whole-chapter links to work.                            #
#        Fixed it so that a whole chapter reference keeps the white space   #
#             after the chapter number.  Before, it didn't put it in,       #
#             making for things like "Matt. 5is good."                      #
# v1.22: Added Words of Morm. as a book option.                             #
# v1.21: Added D. & C. as a possible D&C bookname.                          #
#        Added JS-*H as a possible book name.                               #
#        Added Psalm and convert it correctly.                              #
# v1.2:  Added the code necessary to handle all previous "problem cases",   #
#             Such as these two examples: Gen. 2; 3:5-6:7                   #
# v1.11: Fixed a bug with I John not being recognized.                      #
#        Added a subroutine that prints periods ever 3 seconds to STDERR.   #
# v1.1:  Fixed some book names that weren't listed (D&C variants and others)#
#        Added support for "Ibid" to put the previous book.                 #
#        Added support for abbreviated verses (like Gen 88:105-12)          #
#             This only supports one level of abbreviation. so 105-7        #
#             would still only give you verse 105.                          #
#        Added code that will report how many total references were added   #
#        Fixed a big bug when a 1|2|3 John was in a line, it ignored        #
#             everything after it on the line.  It's the "right" way now.   #
#############################################################################

$make_endnote_links = 0;

make_booklist();									# generate the list of matches

$book_level = "Book";
$quiet = 0;
$web = 0;
$match_partial_references = 0;
my $test_harness = 0;
$printed_warning = 0;

$i = 0;
while ($i < scalar(@ARGV)) {
	if ($ARGV[$i] =~ /^-local$/) {
		$make_endnote_links = 1;
		splice(@ARGV, $i, 1);
		next;
	} elsif ($ARGV[$i] =~ /^-match_partials$/) {
		$match_partial_references = 1;
		splice(@ARGV, $i, 1);
		next;
	} elsif ($ARGV[$i] =~ /^-book_level=(.+)$/) {
		$book_level = $1;
		splice(@ARGV, $i, 1);
		next;
	} elsif ($ARGV[$i] =~ /^-quiet$/) {
		$quiet = 1;
		splice(@ARGV, $i, 1);
		next;
	} elsif ($ARGV[$i] =~ /^-web(?:_links)?$/) {			# For html documents, to do a <a href style link
		$web = 1;
		splice(@ARGV, $i, 1);
		next;
	} elsif ($ARGV[$i] =~ /^-help$/) {
		help();
		exit;
	} elsif ($ARGV[$i] =~ /^-test$/) {
		$test_harness = 1;
		splice(@ARGV, $i, 1);
		next;
	}
	$i++;
}

if ($test_harness) {
	*IN = *DATA;
} else {
	($infile) = $ARGV[0];							# get the input filename from the user
	if (! $infile) {									# ask if not on the command line
	    print STDERR "Input File: ";
	    chomp($infile = <STDIN>);
	}
	open (IN, "<$infile") or die "Can't open input file.\n";
}
if (defined($ARGV[1])) {
	$outfile = $ARGV[1];							# if an output filename is specified, open it
}
if ($outfile) {									#   or just dump it to the screen
    open(STDOUT, ">$outfile") or die "Can't open output file.\n";
}

$number_found = 0;
while (<IN>) {
	$line = $_;
	$new_piece = "";
	$final_line = "";
	
	$line =~ s// /g;											# change that screwy space into a normal space.
	
	$line =~ s/Quad (With Strongs)/Quad/g;
	$line =~ s/<QL.*?Quad\">(.*?)<EL>/$1/g;			# to kill previous references
	$line =~ s/<(?:QL|EN)[^>]*?Field (?:Chapter|Verse)[^>]*?\">(.*?)<EL>/$1/g if ($make_endnote_links);			# to kill previous references
	$line =~ s/<QL.*?quadv4.nfo\">(.*?)<EL>/$1/g;	# to kill previous references
	$line =~ s/<QL.*?,RH\">(.*?)<EL>/$1/g;				# to kill previous references
	$line =~ s/(\d):(\d+\)\(\d+)/$1:$2-$3/g;
	$line =~ s/<WW:Web,\"http:\/\/scriptures.lds.org[^>]+>(.*?)<EL>/$1/g;
	$line =~ s/<WW:Web,\"http:\/\/lds.org\/scriptures[^>]+>(.*?)<EL>/$1/g;
	$line =~ s/<WW:Web,\"http:\/\/www.lds.org\/scriptures[^>]+>(.*?)<EL>/$1/g;

# Try this instead: $line =~ s/\b([0-9]{9,10}\.[0-9]+|[0-9]{9,10})\b/atime($1)/eg if ($ctime);
# This can do a replace on all references in a line, if done right.

	while (($line =~ /^(.*)\b($booklist)([.,]?(?:<IT.?>[.,]?)?\s+)([0-9-]+)(:?)([0-9, -]*)( *.*?)$/o && assign_match_type(1))
			|| ($match_partial_references && $old_bookname && $line =~ /^(.*\()()()([0-9-]+)(:)([0-9, -]*)( *.*?)$/o && assign_match_type(2))
			|| ($match_partial_references && $line =~ /^(.*\D)()()([0-9-]+)(:)([0-9, -]*)( *.*?)$/o && assign_match_type(3))) {
		$head         = $1;								# get the pieces into storage variables
		$book         = $2;
		$book_orig    = $2;
		$space        = $3;
		$chapter      = $4;
		$chapter_orig = $4;
		$found_verse  = $5;
		$verse        = $6;
		$orig_verse   = $6;
		$tail         = $7;
		
		if ($match == 3 && $infile =~ /Great Are the Words of Isaiah/) {
			$old_bookname = "Isaiah";
		}

		if (($verse =~ /^[0-9 -]+,\s+\d+$/o) && ($tail =~ /^:\d+/)) {		# fix ,->; ocr errors
			($new_verse, $new_tail) = ($verse =~ /^([0-9 -]+),(\s+\d+)$/);
			$verse = $new_verse;
			$orig_verse = $verse;
			$tail = ";$new_tail$tail";
		}

		if ($chapter =~ /^(.*)-$/) {				# move a trailing '-' to the beginning of the tail
			$tail = "-$tail";
			$chapter = $1;
			$chapter_orig = $1;
		}

		if ($verse =~ /^(.+?)(\s+)$/) {
			$orig_verse = $1;
			$tail = "$2$tail";
		}
		$verse =~ s/ //g;							# kill any space from the verses
		$verse =~ s//-/g;						# change #150's to dashes

		if ($found_verse eq ":") {	$found_verse = 1;
		} else {							$found_verse = 0; }

		if (! $found_verse) {
			if ($verse) {
				$chapter_orig = "$chapter$orig_verse";
				$chapter = "$chapter,$verse";
				$verse = "";
				$orig_verse = "";
				$chapter = split_number($chapter);
			} else {
				$chapter = split_number($chapter);
			}
		}

		if ($old_bookname && (!defined($book) || !$book)) {
			$bookname = $old_bookname;
		} elsif ($book =~ /^[Ii]bid/) {
			$bookname = $old_bookname;
		} else {
			$bookname = $book;
		}
		$old_bookname = $bookname;
		
				# if I don't look for 1|2|3 John, it would only find John (smallest match)
		if ($bookname =~ /[Jj][ohn]+/) {
			if ($head =~ /(.*?)\b([i|I|1|2|3]+\s*)$/) {
				$head = $1;
				$bookname = $2 . $bookname;
				$book_orig = $2 . $book_orig;
			}
		} elsif ($bookname =~ /Mormon/) {
			if ($head =~ /(.*?)\b(Words of \s*)$/) {
				$head = $1;
				$bookname = $2 . $bookname;
				$book_orig = $2 . $book_orig;
			}
		} elsif ($bookname =~ /Section/i) {
			if ($head =~ /(.*?)\b(Doctrine \S+ Covenants\s*)$/) {
				$head = $1;
				$bookname = $2 . $bookname;
				$book_orig = $2 . $book_orig;
			}
		} elsif ($bookname =~ /Matthew/i) {
			if ($head =~ /(.*?)\b(Joseph Smith[ -])$/) {
				$head = $1;
				$bookname = $2 . $bookname;
				$book_orig = $2 . $book_orig;
			}
		}

		if ($bookname !~ /^i/) {
			$bookname = ucfirst($bookname);
		} else {
			$bookname =~ s/^([iv]+) /\U$1/;
		}

		($bookname, $chapter, $verse) = convert_book($bookname, $chapter, $verse);						# call subroutine to convert short->long names

		if ($bookname ne "History of the Church") {
			if ($chapter =~ /^\d+$/ && defined($chapters{$bookname}) && $chapter > $chapters{$bookname}) {		# Ignore complex chapters like Mark 3-6 for now...
				$extra = "\n";
				$extra = "" if ($printed_warning);
				print STDERR "${extra}Warning: Out of range chapter for the given book: $bookname $chapter_orig\n$extra" if (!$quiet);
				$printed_warning++;
			} elsif (!defined($chapters{$bookname})) {
				$extra = "\n";
				$extra = "" if ($printed_warning);
				print STDERR "${extra}Warning: No chapter range defined for book: $bookname\n$extra" if (!$quiet);
				$printed_warning++;
			}
		}

		if ($verse && $tail !~ /^:([0-9]+)(.*)/) {
			$verse = split_number($verse);
		}
		
		
												# build up the query string
		if ($web) {							# http://scriptures.lds.org/en/dc/88/22-24#22
			$new_piece = "<a href=\"http://scriptures.lds.org/en/$bookname/$chapter/";
			$new_piece .= "$verse#$verse" if ($verse);
			$new_piece .= "\">$book_orig$space$chapter_orig";
			$new_piece .= ":$orig_verse" if ($verse);
			$new_piece .= "</a>";
		} else {
			$new_piece = "<QL:Query,\"[Headings $book_level,$bookname] & [Field Chapter:$chapter]";
			$new_piece .= " & [Field Verse:$verse]" if ($verse);
			$new_piece .= "\",RH,\"quadv4.nfo\">$book_orig$space$chapter_orig";
			$new_piece .= ":" if ($orig_verse && $verse);
			$new_piece .= "$orig_verse" if ($verse);
			$new_piece .= "<EL>";
		}

		if ($orig_verse =~ /^\s+$/) {			# add the space after chapter (if without a verse)
			$new_piece .= "$orig_verse";		# otherwise, "Ex 4 is" becomes "Ex 4is"
		}

		if ($orig_verse =~ /^([0-9]+)[-]([0-9]+)/o) {	# this checks for 4:5-5:6 type references
			$temp1 = $1;
			$temp2 = $2;
			if ($tail =~ /^:([0-9]+)(.*)/) {
				$tmp1 = $1;
				$tail = $2;
				$new_piece = "<QL:Query,\"[Headings $book_level,$bookname] & (([Field Chapter:$chapter] & [Field Verse:>=$temp1])";
				while ($chapter < $temp2-1) {
					$chapter++;
					$new_piece .= " | [Field Chapter:$chapter]";
				}
				$new_piece .= " | ([Field Chapter:$temp2] & [Field Verse:<=$1]))\",RH,\"quadv4.nfo\">$book_orig$space$chapter_orig:$orig_verse:$1<EL>";
			}
		}
		
		if ($orig_verse =~ /^-(\d+)(\s*)$/) {	# Added to do multi- whole-chapter links
			$new_piece = "<QL:Query,\"[Headings $book_level,$bookname] & [Field Chapter:$chapter";
			$this_chapter = $chapter + 1;
			while ($this_chapter <= $1) {
				$new_piece .= "|$this_chapter";
				$this_chapter++;
			}
			$new_piece .= "]\",RH,\"quadv4.nfo\">$book_orig$space$chapter-$1$2<EL>";
		}

		split_extra_chapters();							# check for multi-chapter references (John 1:2; 3:4)

		$new_piece =~ s/<QL:Query,("[^\"]+"),RH,"quadv4.nfo">(.*?)<EL>/<EN:Query,$1,5,2.5,"$2">$2<EL>/g if (!$web && $make_endnote_links);
#		$new_piece =~ s/quadv4.nfo/History of the Church.nfo/g if ($bookname eq "History of the Church");

		$final_line = $new_piece . $final_line;	# rebuild the line, in order (from the end to the front)
		$line = $head;										# shift over the next piece of the line over to get checked for a reference
		$number_found++;
		$verse = "";
	}
	if ($new_piece) {
		if ($web) {
			$final_line =~ s/ <\/a>/<\/a> /g;
			$final_line =~ s/,<\/a>/<\/a>,/g;
		} else {
			$final_line =~ s/ <EL>/<EL> /g;				# shift link ends before spaces or commas
			$final_line =~ s/,<EL>/<EL>,/g;
		}
		print STDOUT "$head$final_line\n";			# if there was a reference, print out the line
	} else {
		print STDOUT $line;								# if not, print out the unmodified line
	}
	tick();
}
print STDERR "\n$number_found references were added";
print STDERR " in ", ($sec - $start_time) , " seconds." if ($sec > $start_time);
print STDERR "\n";

#############################################################################
# Subroutines
#############################################################################

#############################################################################
sub split_extra_chapters {
	$new_tail = "";								# make some temp variables
	$next_tail = $tail;
	while ($next_tail =~ /^;(\s*)([0-9-]+)(:?)([0-9, -]*)( *.*)$/o) {
		$this_piece   = "";
		$more_space   = $1;						# store the chunks into vars
		$chapter      = $2;
		$chapter_orig = $2;
		$found_verse  = $3;
		$verse        = $4;
		$orig_verse   = $4;
		$new_tail     = $5;

		if ($chapter =~ /^(.*)-$/) {
			$new_tail = "-$new_tail";
			$chapter = $1;
		}

		if ($verse =~ /^(.+?)(\s+)$/) {
			$orig_verse = $1;
			$new_tail = "$2$new_tail";
		}
		$verse =~ s/ //g;						# kill any space from the verses

		if ($found_verse eq ":") {
			$found_verse = 1;
		} else {
			$found_verse = 0;
		}

		if (! $found_verse) {
			if ($verse) {
				$chapter_orig = "$chapter$orig_verse";
				$chapter      = "$chapter,$verse";
				$verse        = "";
				$orig_verse   = "";
				$chapter      = split_number($chapter);
			} else {
				$chapter      = split_number($chapter);
			}
		}

		$verse = split_number($verse) if ($new_tail !~ /^:([0-9]+)(.*)/);							# split the verses up into pieces

		$this_piece .= ";$more_space<QL:Query,\"[Headings $book_level,$bookname] & [Field Chapter:$chapter]";
		if ($verse) {
			$this_piece .= " & [Field Verse:$verse]";
		}
		$this_piece .= "\",RH,\"quadv4.nfo\">$chapter_orig";
		if ($orig_verse) { $this_piece .= ":"; }
		$this_piece .= "$orig_verse<EL>";

		if ($orig_verse =~ /^([0-9]+)-([0-9]+)/) {	# this checks for 4:5-6:7 type references
			$temp1 = $1;
			$temp2 = $2;
			if ($new_tail =~ /^:([0-9]+)(.*)/) {
				$tmp1 = $1;
				$new_tail = $2;
				$this_piece = ";$more_space<QL:Query,\"[Headings $book_level,$bookname] & (([Field Chapter:$chapter] & [Field Verse:>=$temp1])";
				while ($chapter < $temp2-1) {
					$chapter++;
					$this_piece .= " | [Field Chapter:$chapter]";
				}
				$this_piece .= " | ([Field Chapter:$temp2] & [Field Verse:<=$1]))\",RH,\"quadv4.nfo\">$book_orig$space$chapter_orig:$orig_verse:$1<EL>";
			}
		}

		$new_piece = $new_piece . $this_piece;
		$next_tail = $new_tail;					# shift to the next set of chapter:verse references
		$number_found++;
	}
	if ($new_piece) {
		$new_piece .= $next_tail;			# stick the pieces together
	} else {
		$new_piece .= $tail;					# otherwise, restore the line end
	}
}

#############################################################################
sub split_number {
	my ($number) = @_;
	$new_num = "";
	if ($number =~ /(-|,)/) {
		$number =~ s/,+/,/g;
		@nums = split(/,/, $number);
		$i = 0;
		while ($i < scalar(@nums)) {
			if ($nums[$i] =~ /(\d+)[-](\d+)/) {
				$new_num  = $new_num . "|" . $1;
				$lowernum = $1 + 1;
				$uppernum = $2;
				if ($lowernum > $uppernum) {
					if ($lowernum =~ /^(.)/) {
						$first_char = $1;
					}
					$uppernum = $first_char . $uppernum;
				}
				while ($lowernum <= $uppernum) {
					$new_num = $new_num . "|$lowernum";
					$lowernum++;
					if ($lowernum > 150) {
						print STDERR "\nquad.pl: ERROR: A chapter or verse range is > 150, which shouldn't happen. ($number)\n\n  $line\n";
						return($number);
					}
				}
			} else {
				$new_num = $new_num . "|" . $nums[$i];
			}
			$i++;
		}
	} else { $new_num = $number; }
	$new_num =~ s/^\|//;
	return($new_num);
}

#############################################################################
sub tick
{
	$sec = time;
	if (! $trigger_sec) {
		$tick_offset = 3;
		$trigger_sec = $sec + $tick_offset;
		$start_time  = $sec;
	}
	while ($trigger_sec <= $sec) {
		print STDERR ".";
		$trigger_sec = $trigger_sec + $tick_offset;
	}
}

#############################################################################
sub assign_match_type {
	$match = $_[0];
	return(1);
}

#############################################################################
# This is the list of all books that will be considered a reference to the scriptures.
sub make_booklist {
	@longlist = ('Gen',			'Genesis',
			 '3 Jn',			'3 John',				'III Jn',		'III John',
			 '2 Jn',			'2 John',				'II Jn',			'II John',
			 '1 Jn',			'1 John',				'I Jn',			'I John',
			 'Ex',			'Exodus',				'Exod',			'Exo',
			 'Lev',			'Leviticus',
			 'Num',			'Numbers',				'Numb',
			 'Deut',			'Deuteronomy',
			 'Josh',			'Joshua',
			 'Judg',			'Judges',
			 'Ruth',
			 '1 Sam',		'1 Samuel',				'I Sam',			'I Samuel',
			 '2 Sam',		'2 Samuel',				'II Sam',		'II Samuel',
			 '1 Kgs',		'1 Kings',				'I Kgs',			'I Kings',
			 '2 Kgs',		'2 Kings',				'II Kgs',		'II Kings',
			 '1 Chr',		'1 Chronicles',		'I Chr',			'I Chronicles',	'1 Chron',
			 '2 Chr',		'2 Chronicles',		'II Chr',		'II Chronicles',	'2 Chron',
			 'Ezra',
			 'Neh',			'Nehemiah',
			 'Esth',			'Esther',
			 'Job',
			 'Ps',			'Psalms',				'Psalm',
			 'Prov',			'Proverbs',
			 'Eccl',			'Ecclesiastes',
			 'Song',			'Song of Solomon',
			 'Isa',			'Isaiah',
			 'Jer',			'Jeremiah',
			 'Lam',			'Lamentations',
			 'Ezek',			'Ezekiel',
			 'Dan',			'Daniel',
			 'Hosea',		'Hos',
			 'Joel',
			 'Amos',
			 'Obad',			'Obadiah',
			 'Jonah',
			 'Micah',
			 'Nahum',
			 'Hab',			'Habakkuk',
			 'Zeph',			'Zephaniah',
			 'Hag',			'Haggai',
			 'Zech',			'Zechariah',
			 'Mal',			'Malachi',

			 'Matt',			'Matthew',
			 'Mark',
			 'Luke',
			 'John',
			 'Acts',			'Acts',
			 'Rom',			'Romans',
			 '1 Cor',		'1 Corinthians',		'I Cor',		'I Corinthians',
			 '2 Cor',		'2 Corinthians',		'II Cor',		'II Corinthians',
			 'Gal',			'Galatians',
			 'Eph',			'Ephesians',
			 'Philip',		'Philippians',
			 'Col',			'Colossians',
			 '1 Thes',		'1 Thessalonians',	'I Thes',		'I Thessalonians',
			 '2 Thes',		'2 Thessalonians',	'II Thes',		'II Thessalonians',
			 '1 Tim',		'1 Timothy',			'I Tim',			'I Timothy',
			 '2 Tim',		'2 Timothy',			'II Tim',		'II Timothy',
			 'Titus',
			 'Philem',		'Philemon',
			 'Heb',			'Hebrews',
			 'James',
			 '1 Pet',		'1 Peter',				'I Pet',			'I Peter',
			 '2 Pet',		'2 Peter',				'II Pet',		'II Peter',
			 'Jude',
			 'Rev',			'Revelation',			'Revelations',

			 '1 Ne',			'1 Nephi',				'I Ne',			'I Nephi',
			 '2 Ne',			'2 Nephi',				'II Ne',			'II Nephi',
			 'Jacob',
			 'Enos',
			 'Jarom',
			 'Omni',
			 'W of M',		'Words of Mormon',		'Words of Morm',
			 'Mosiah',
			 'Alma',
			 'Hel',			'Hela',					'Helaman',
			 '3 Ne',			'3 Nephi',				'III Ne',		'III Nephi',
			 '4 Ne',			'4 Nephi',				'IV Ne',			'IV Nephi',
			 'Morm',			'Mormon',
			 'Ether',		'Eth',
			 'Moro',			'Moroni',

			 'D & C',		'Doctrine and Covenants',	'Doctrine & Covenants',	'D. & C.',	'Doc. & Cov.',	'Doc. and Cov.', 'Section', 'D\.? & C\.?,? Sec.?', 'D\.? and C\.?,? Sec.?', 'D&C,? Section',
			 'OD',			'Official Declaration',
			 'OD1',			'Official Declaration 1',
			 'OD2',			'Official Declaration 2',

			 'Moses',
			 'Abr',			'Abraham',
			 'JS-M',			'JSM',					'Joseph Smith Matthew',		'JS Matt',	'Joseph SmithMatthew',		'Joseph Smith-Matthew',
			 'JSH',			'Joseph Smith-History',	'Joseph Smith History',		'JS-H',			'JS--H',	'JSH',		'Joseph SmithHistory',		'JSHistory',	'JS-History',
			 'A of F',		'Articles of Faith',
#			 'History of the Church', 'HC',
#			 'Ibid.,',		'Ibid,'
			 );


	$booklist = join('|', @longlist);				# make it all one text variable with | separating
	$booklist =~ s/([A-Z])/[$1\l$1]/g;				# add uppercase/lowercase code
	$booklist =~ s/\s+/\\s*/g;							# make all space optional
	$booklist = qr($booklist);							# pre-compile it for speed!

#	Book statistics
#	Books               -> Chapters  (per book)
	$chapters{"Genesis"}			= 50;
	$chapters{"Exodus"}			= 40;
	$chapters{"Leviticus"}		= 27;
	$chapters{"Numbers"}			= 36;
	$chapters{"Deuteronomy"}	= 34;
	$chapters{"Joshua"}			= 24;
	$chapters{"Judges"}			= 21;
	$chapters{"Ruth"}				= 4;
	$chapters{"1 Samuel"}		= 31;
	$chapters{"2 Samuel"}		= 24;
	$chapters{"1 Kings"}			= 22;
	$chapters{"2 Kings"}			= 25;
	$chapters{"1 Chronicles"}	= 29;
	$chapters{"2 Chronicles"}	= 36;
	$chapters{"Ezra"}				= 10;
	$chapters{"Nehemiah"}		= 13;
	$chapters{"Esther"}			= 10;
	$chapters{"Job"}				= 42;
	$chapters{"Psalms"}			= 150;
	$chapters{"Proverbs"}		= 31;
	$chapters{"Ecclesiastes"}	= 12;
	$chapters{"Song of Solomon"}	= 8;
	$chapters{"Isaiah"}			= 66;
	$chapters{"Jeremiah"}		= 52;
	$chapters{"Lamentations"}	= 5;
	$chapters{"Ezekiel"}			= 48;
	$chapters{"Daniel"}			= 12;
	$chapters{"Hosea"}			= 14;
	$chapters{"Joel"}				= 3;
	$chapters{"Amos"}				= 9;
	$chapters{"Obadiah"}			= 1;
	$chapters{"Jonah"}			= 4;
	$chapters{"Micah"}			= 7;
	$chapters{"Nahum"}			= 3;
	$chapters{"Habakkuk"}		= 3;
	$chapters{"Zephaniah"}		= 3;
	$chapters{"Haggai"}			= 2;
	$chapters{"Zechariah"}		= 14;
	$chapters{"Malachi"}			= 4;
	
	$chapters{"Matthew"}			= 28;
	$chapters{"Mark"}				= 16;
	$chapters{"Luke"}				= 24;
	$chapters{"John"}				= 21;
	$chapters{"Acts"}				= 28;
	$chapters{"Romans"}			= 16;
	$chapters{"1 Corinthians"}	= 16;
	$chapters{"2 Corinthians"}	= 13;
	$chapters{"Galatians"}		= 6;
	$chapters{"Ephesians"}		= 6;
	$chapters{"Philippians"}	= 4;
	$chapters{"Colossians"}		= 4;
	$chapters{"1 Thessalonians"}	= 5;
	$chapters{"2 Thessalonians"}	= 3;
	$chapters{"1 Timothy"}		= 6;
	$chapters{"2 Timothy"}		= 4;
	$chapters{"Titus"}			= 3;
	$chapters{"Philemon"}		= 1;
	$chapters{"Hebrews"}			= 13;
	$chapters{"James"}			= 5;
	$chapters{"1 Peter"}			= 5;
	$chapters{"2 Peter"}			= 3;
	$chapters{"1 John"}			= 5;
	$chapters{"2 John"}			= 1;
	$chapters{"3 John"}			= 1;
	$chapters{"Jude"}				= 1;
	$chapters{"Revelation"}		= 22;

	$chapters{"1 Nephi"}			= 22;
	$chapters{"2 Nephi"}			= 33;
	$chapters{"Jacob"}			= 7;
	$chapters{"Enos"}				= 1;
	$chapters{"Jarom"}			= 1;
	$chapters{"Omni"}				= 1;
	$chapters{"Words of Mormon"}	= 1;
	$chapters{"Mosiah"}			= 29;
	$chapters{"Alma"}				= 63;
	$chapters{"Helaman"}			= 16;
	$chapters{"3 Nephi"}			= 30;
	$chapters{"4 Nephi"}			= 1;
	$chapters{"Mormon"}			= 9;
	$chapters{"Ether"}			= 15;
	$chapters{"Moroni"}			= 10;

	$chapters{"'D&C'"}			= 138;

	$chapters{"Moses"}			= 8;
	$chapters{"Abraham"}			= 5;
	$chapters{"Joseph Smith Matthew"}	= 1;
	$chapters{"Joseph Smith History"}	= 1;
	$chapters{"Articles of Faith"}		= 1;
}


#############################################################################
# This subroutine converts the name of any book to the given full name of the book.
# The \A..\Z ensures that it's a full match.  Ex.: 'JSMatt' won't become 'Matthew'
sub convert_book {
	my $bookname = $_[0];
	my $chapter  = $_[1];
	my $verse    = $_[2];

	if (defined($chapters{$bookname})) {
		if ($chapters{$bookname} == 1) {
			if ((!defined($verse) || !$verse) && $chapter) {
				$verse = $chapter;
			}
			$chapter = 1;
		}

		return($bookname, $chapter, $verse);
	}
	
	if    ($bookname =~ /^[Gg]en$/)						{ $bookname = "Genesis"; }
	elsif ($bookname =~ /^[Ee]x$/)						{ $bookname = "Exodus"; }
	elsif ($bookname =~ /^[Ee]xod$/)						{ $bookname = "Exodus"; }
	elsif ($bookname =~ /^[Ll]ev$/)						{ $bookname = "Leviticus"; }
	elsif ($bookname =~ /^[Nn]um$/)						{ $bookname = "Numbers"; }
	elsif ($bookname =~ /^[Dd]eut$/)						{ $bookname = "Deuteronomy"; }
	elsif ($bookname =~ /^[Jj]osh$/)						{ $bookname = "Joshua"; }
	elsif ($bookname =~ /^[Jj]udg$/)						{ $bookname = "Judges"; }
	elsif ($bookname =~ /^1\s*[Ss]am$/)					{ $bookname = "1 Samuel"; }
	elsif ($bookname =~ /^I\s*[Ss]am$/)					{ $bookname = "1 Samuel"; }
	elsif ($bookname =~ /^I\s*[Ss]amuel$/)				{ $bookname = "1 Samuel"; }
	elsif ($bookname =~ /^1\s*[Ss]amuel$/)				{ $bookname = "1 Samuel"; }
	elsif ($bookname =~ /^2\s*[Ss]am$/)					{ $bookname = "2 Samuel"; }
	elsif ($bookname =~ /^II\s*[Ss]am$/)				{ $bookname = "2 Samuel"; }
	elsif ($bookname =~ /^II\s*[Ss]amuel$/)			{ $bookname = "2 Samuel"; }
	elsif ($bookname =~ /^2\s*[Ss]amuel$/)				{ $bookname = "2 Samuel"; }
	elsif ($bookname =~ /^1\s*[Kk]gs$/)					{ $bookname = "1 Kings"; }
	elsif ($bookname =~ /^I\s*[Kk]gs$/)					{ $bookname = "1 Kings"; }
	elsif ($bookname =~ /^I\s*[Kk]ings$/)				{ $bookname = "1 Kings"; }
	elsif ($bookname =~ /^1\s*[Kk]ings$/)				{ $bookname = "1 Kings"; }
	elsif ($bookname =~ /^2\s*[Kk]gs$/)					{ $bookname = "2 Kings"; }
	elsif ($bookname =~ /^II\s*[Kk]gs$/)				{ $bookname = "2 Kings"; }
	elsif ($bookname =~ /^II\s*[Kk]ings$/)				{ $bookname = "2 Kings"; }
	elsif ($bookname =~ /^2\s*[Kk]ings$/)				{ $bookname = "2 Kings"; }
	elsif ($bookname =~ /^1\s*[Cc]hr$/)					{ $bookname = "1 Chronicles"; }
	elsif ($bookname =~ /^I\s*[Cc]hr$/)					{ $bookname = "1 Chronicles"; }
	elsif ($bookname =~ /^1\s*[Cc]hron$/)				{ $bookname = "1 Chronicles"; }
	elsif ($bookname =~ /^I\s*[Cc]hron$/)				{ $bookname = "1 Chronicles"; }
	elsif ($bookname =~ /^I\s*[Cc]hronicles$/)		{ $bookname = "1 Chronicles"; }
	elsif ($bookname =~ /^1\s*[Cc]hronicles$/)		{ $bookname = "1 Chronicles"; }
	elsif ($bookname =~ /^2\s*[Cc]hr$/)					{ $bookname = "2 Chronicles"; }
	elsif ($bookname =~ /^II\s*[Cc]hr$/)				{ $bookname = "2 Chronicles"; }
	elsif ($bookname =~ /^II\s*[Cc]hronicles$/)		{ $bookname = "2 Chronicles"; }
	elsif ($bookname =~ /^2\s*[Cc]hronicles$/)		{ $bookname = "2 Chronicles"; }
	elsif ($bookname =~ /^[Nn]eh$/)						{ $bookname = "Nehemiah"; }
	elsif ($bookname =~ /^[Ee]sth$/)						{ $bookname = "Esther"; }
	elsif ($bookname =~ /^[Pp]s$/)						{ $bookname = "Psalms"; }
	elsif ($bookname =~ /^[Pp]salm$/)					{ $bookname = "Psalms"; }
	elsif ($bookname =~ /^[Pp]rov$/)						{ $bookname = "Proverbs"; }
	elsif ($bookname =~ /^[Ee]ccl$/)						{ $bookname = "Ecclesiastes"; }
	elsif ($bookname =~ /^[Ss]ong$/)						{ $bookname = "Song of Solomon"; }
	elsif ($bookname =~ /^[Ii]sa$/)						{ $bookname = "Isaiah"; }
	elsif ($bookname =~ /^[Jj]er$/)						{ $bookname = "Jeremiah"; }
	elsif ($bookname =~ /^[Ll]am$/)						{ $bookname = "Lamentations"; }
	elsif ($bookname =~ /^[Ee]zek$/)						{ $bookname = "Ezekiel"; }
	elsif ($bookname =~ /^[Dd]an$/)						{ $bookname = "Daniel"; }
	elsif ($bookname =~ /^[Oo]bad$/)						{ $bookname = "Obadiah"; }
	elsif ($bookname =~ /^[Hh]ab$/)						{ $bookname = "Habakkuk"; }
	elsif ($bookname =~ /^[Zz]eph$/)						{ $bookname = "Zephaniah"; }
	elsif ($bookname =~ /^[Hh]ag$/)						{ $bookname = "Haggai"; }
	elsif ($bookname =~ /^[Zz]ech$/)						{ $bookname = "Zechariah"; }
	elsif ($bookname =~ /^[Mm]al$/)						{ $bookname = "Malachi"; }

	elsif ($bookname =~ /^[Mm]att$/)						{ $bookname = "Matthew"; }
	elsif ($bookname =~ /^[Rr]om$/)						{ $bookname = "Romans"; }
	elsif ($bookname =~ /^1\s*[Cc]or$/)					{ $bookname = "1 Corinthians"; }
	elsif ($bookname =~ /^I\s*[Cc]or$/)					{ $bookname = "1 Corinthians"; }
	elsif ($bookname =~ /^I\s*[Cc]orinthians$/)		{ $bookname = "1 Corinthians"; }
	elsif ($bookname =~ /^1\s*[Cc]orinthians$/)		{ $bookname = "1 Corinthians"; }
	elsif ($bookname =~ /^2\s*[Cc]or$/)					{ $bookname = "2 Corinthians"; }
	elsif ($bookname =~ /^II\s*[Cc]or$/)				{ $bookname = "2 Corinthians"; }
	elsif ($bookname =~ /^II\s*[Cc]orinthians$/)		{ $bookname = "2 Corinthians"; }
	elsif ($bookname =~ /^2\s*[Cc]orinthians$/)		{ $bookname = "2 Corinthians"; }
	elsif ($bookname =~ /^[Gg]al$/)						{ $bookname = "Galatians"; }
	elsif ($bookname =~ /^[Ee]ph$/)						{ $bookname = "Ephesians"; }
	elsif ($bookname =~ /^[Pp]hilip$/)					{ $bookname = "Philippians"; }
	elsif ($bookname =~ /^[Cc]ol$/)						{ $bookname = "Colossians"; }
	elsif ($bookname =~ /^1\s*[Tt]hes$/)				{ $bookname = "1 Thessalonians"; }
	elsif ($bookname =~ /^I\s*[Tt]hes$/)				{ $bookname = "1 Thessalonians"; }
	elsif ($bookname =~ /^I\s*[Tt]hessalonians$/)	{ $bookname = "1 Thessalonians"; }
	elsif ($bookname =~ /^1\s*[Tt]hessalonians$/)	{ $bookname = "1 Thessalonians"; }
	elsif ($bookname =~ /^2\s*[Tt]hes$/)				{ $bookname = "2 Thessalonians"; }
	elsif ($bookname =~ /^II\s*[Tt]hes$/)				{ $bookname = "2 Thessalonians"; }
	elsif ($bookname =~ /^II\s*[Tt]hessalonians$/)	{ $bookname = "2 Thessalonians"; }
	elsif ($bookname =~ /^2\s*[Tt]hessalonians$/)	{ $bookname = "2 Thessalonians"; }
	elsif ($bookname =~ /^1\s*[Tt]im$/)					{ $bookname = "1 Timothy"; }
	elsif ($bookname =~ /^I\s*[Tt]im$/)					{ $bookname = "1 Timothy"; }
	elsif ($bookname =~ /^I\s*[Tt]imothy$/)			{ $bookname = "1 Timothy"; }
	elsif ($bookname =~ /^1\s*[Tt]imothy$/)			{ $bookname = "1 Timothy"; }
	elsif ($bookname =~ /^2\s*[Tt]im$/)					{ $bookname = "2 Timothy"; }
	elsif ($bookname =~ /^II\s*[Tt]im$/)				{ $bookname = "2 Timothy"; }
	elsif ($bookname =~ /^II\s*[Tt]imothy$/)			{ $bookname = "2 Timothy"; }
	elsif ($bookname =~ /^2\s*[Tt]imothy$/)			{ $bookname = "2 Timothy"; }
	elsif ($bookname =~ /^[Pp]hilem$/)					{ $bookname = "Philemon"; }
	elsif ($bookname =~ /^[Hh]eb$/)						{ $bookname = "Hebrews"; }
	elsif ($bookname =~ /^1\s*[Pp]et$/)					{ $bookname = "1 Peter"; }
	elsif ($bookname =~ /^I\s*[Pp]et$/)					{ $bookname = "1 Peter"; }
	elsif ($bookname =~ /^I\s*[Pp]eter$/)				{ $bookname = "1 Peter"; }
	elsif ($bookname =~ /^1\s*[Pp]eter$/)				{ $bookname = "1 Peter"; }
	elsif ($bookname =~ /^2\s*[Pp]et$/)					{ $bookname = "2 Peter"; }
	elsif ($bookname =~ /^II\s*[Pp]et$/)				{ $bookname = "2 Peter"; }
	elsif ($bookname =~ /^II\s*[Pp]eter$/)				{ $bookname = "2 Peter"; }
	elsif ($bookname =~ /^2\s*[Pp]eter$/)				{ $bookname = "2 Peter"; }
	elsif ($bookname =~ /^1\s*[Jj]n$/)					{ $bookname = "1 John"; }
	elsif ($bookname =~ /^I\s*[Jj]n$/)					{ $bookname = "1 John"; }
	elsif ($bookname =~ /^I\s*[Jj]ohn$/)				{ $bookname = "1 John"; }
	elsif ($bookname =~ /^1\s*[Jj]ohn$/)				{ $bookname = "1 John"; }
	elsif ($bookname =~ /^2\s*[Jj]n$/)					{ $bookname = "2 John"; }
	elsif ($bookname =~ /^II\s*[Jj]n$/)					{ $bookname = "2 John"; }
	elsif ($bookname =~ /^II\s*[Jj]ohn$/)				{ $bookname = "2 John"; }
	elsif ($bookname =~ /^2\s*[Jj]ohn$/)				{ $bookname = "2 John"; }
	elsif ($bookname =~ /^3\s*[Jj]n$/)					{ $bookname = "3 John"; }
	elsif ($bookname =~ /^III\s*[Jj]n$/)				{ $bookname = "3 John"; }
	elsif ($bookname =~ /^III\s*[Jj]ohn$/)				{ $bookname = "3 John"; }
	elsif ($bookname =~ /^3\s*[Jj]ohn$/)				{ $bookname = "3 John"; }
	elsif ($bookname =~ /^[Rr]ev$/)						{ $bookname = "Revelation"; }
	elsif ($bookname =~ /^[Rr]evelations$/)			{ $bookname = "Revelation"; }

	elsif ($bookname =~ /^1\s*[Nn]e$/)					{ $bookname = "1 Nephi"; }
	elsif ($bookname =~ /^I\s*[Nn]e$/)					{ $bookname = "1 Nephi"; }
	elsif ($bookname =~ /^I\s*[Nn]ephi$/)				{ $bookname = "1 Nephi"; }
	elsif ($bookname =~ /^1\s*[Nn]ephi$/)				{ $bookname = "1 Nephi"; }
	elsif ($bookname =~ /^2\s*[Nn]e$/)					{ $bookname = "2 Nephi"; }
	elsif ($bookname =~ /^II\s*[Nn]e$/)					{ $bookname = "2 Nephi"; }
	elsif ($bookname =~ /^II\s*[Nn]ephi$/)				{ $bookname = "2 Nephi"; }
	elsif ($bookname =~ /^2\s*[Nn]ephi$/)				{ $bookname = "2 Nephi"; }
	elsif ($bookname =~ /^[Ww]\s*of\s*[Mm]$/)			{ $bookname = "Words of Mormon"; }
	elsif ($bookname =~ /^[Ww]ords\s*of\s*[Mm]orm$/)	{ $bookname = "Words of Mormon"; }
	elsif ($bookname =~ /^[Hh]ela?$/)					{ $bookname = "Helaman"; }
	elsif ($bookname =~ /^3\s*[Nn]e$/)					{ $bookname = "3 Nephi"; }
	elsif ($bookname =~ /^III\s*[Nn]e$/)				{ $bookname = "3 Nephi"; }
	elsif ($bookname =~ /^III\s*[Nn]ephi$/)			{ $bookname = "3 Nephi"; }
	elsif ($bookname =~ /^3\s*[Nn]ephi$/)				{ $bookname = "3 Nephi"; }
	elsif ($bookname =~ /^4\s*[Nn]e$/)					{ $bookname = "4 Nephi"; }
	elsif ($bookname =~ /^IV\s*[Nn]e$/)					{ $bookname = "4 Nephi"; }
	elsif ($bookname =~ /^IV\s*[Nn]ephi$/)				{ $bookname = "4 Nephi"; }
	elsif ($bookname =~ /^4\s*[Nn]ephi$/)				{ $bookname = "4 Nephi"; }
	elsif ($bookname =~ /^[Mm]orm$/)						{ $bookname = "Mormon"; }
	elsif ($bookname =~ /^[Ee]th$/)						{ $bookname = "Ether"; }
	elsif ($bookname =~ /^[Mm]oro$/)						{ $bookname = "Moroni"; }

	elsif ($bookname =~ /^[Dd]octrine & [Cc]ovenants$/)	{ $bookname = "'D&C'"; }
	elsif ($bookname =~ /^[Dd]octrine and [Cc]ovenants$/)	{ $bookname = "'D&C'"; }
	elsif ($bookname =~ /^[Dd]\.?\s*&\s*[Cc]\.?$/)			{ $bookname = "'D&C'"; }
	elsif ($bookname =~ /^[Dd]oc\.?\s*(&|and)\s*[Cc]ov\.?$/)		{ $bookname = "'D&C'"; }
	elsif ($bookname =~ /^D\S*\s*(&|and)\s*C\S*\s+[Ss]ection$/)	{ $bookname = "'D&C'"; }
	elsif ($bookname =~ /^D\.?\s*and\s*C\.?,? Sec\S*$/)	{ $bookname = "'D&C'"; }
	elsif ($bookname =~ /^D\.?\s*\&\s*C\.?,? Sec\S*$/)		{ $bookname = "'D&C'"; }
	elsif ($bookname =~ /^Section$/)								{ $bookname = "'D&C'"; }
	elsif ($bookname =~ /^[Oo][Dd]\s*1$/)						{ $bookname = "Official Declaration 1"; }
	elsif ($bookname =~ /^[Oo][Dd]\s*2$/)						{ $bookname = "Official Declaration 2"; }

	elsif ($bookname =~ /^[Aa]br$/)						{ $bookname = "Abraham"; }
	elsif ($bookname =~ /^[Jj][Ss]-*[Mm]$/)			{ $bookname = "Joseph Smith Matthew"; }
	elsif ($bookname =~ /^[Jj][Ss]\s*[Mm]att$/)		{ $bookname = "Joseph Smith Matthew"; }
   elsif ($bookname =~ /^[Jj]oseph [Ss]mith[-]+Matthew$/)	{ $bookname = "Joseph Smith Matthew"; }
	elsif ($bookname =~ /^([Jj][Ss]-*[Hh]|[Jj][Ss][Hh]|[Jj]oseph [Ss]mith[-]+[Hh]istory|JS[-][Hh]istory)$/) {
		$bookname = "Joseph Smith History"; }
	elsif ($bookname =~ /^[Aa]\.?\s*of\s*[Ff]\.?$/)			{ $bookname = "Articles of Faith"; }
	elsif ($bookname =~ /^([Aa][Ff]|Articles of Faith)$/)	{ $bookname = "Articles of Faith"; }

	if ($bookname =~ /^(Jude|Philemon|Enos|4 Nephi|2 John|3 John|Omni|Words of Mormon|Articles of Faith|Joseph Smith History)$/) {
		if ($chapter =~ /\|/ || $chapter > 1) {
			$verse = $chapter;
		}
		$chapter = 1;
	}
	
#	elsif ($bookname =~ /^HC$/)   { $bookname = "History of the Church"; }
	
	return($bookname, $chapter, $verse);
}

#############################################################################
sub help {
	print "Help.\n";
}

__DATA__

<RD>Test.
<RD>Test John 3:5
<RD>Test Matt. 5:4-6
<RD>Test gen 4:3-7, 15-16, 19
<RD>Test I John 2:5-10; exodus 1:1
<RD>Test luke 5:2-6:4
<RD>Test Mark 2; 3.  Gen 2:3; 3; 4:3-9
<RD>Test I think Matt. 5 is good.
<RD>Test D&C 88:105-12; Deut 10:10-5
<RD>Test Job 1-3 is great.
<RD>Test Mosiah 8-11, 13, 17 is great
<RD>Test I like it (John 5:3, 6:1)
<RD>Test Omni 3
<RD>Test Section 3
<RD>John<IT> 20:34-22:14
<RD>Luke, 182-83
<RD>
